/*
 * Copyright (c) 2019 AlexIIL
 *
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at https://mozilla.org/MPL/2.0/.
 */
package alexiil.mc.lib.multipart.api;

import javax.annotation.Nullable;
import net.minecraft.class_1922;
import net.minecraft.class_1937;
import net.minecraft.class_2338;
import net.minecraft.class_2586;
import alexiil.mc.lib.multipart.api.MultipartContainer.MultipartCreator;
import alexiil.mc.lib.multipart.api.MultipartContainer.PartOffer;
import alexiil.mc.lib.multipart.impl.MultipartUtilImpl;

/** Contains various utilities for creating, accessing, or interacting with {@link MultipartContainer}'s in a
 * {@link class_1937}. */
public final class MultipartUtil {
    private MultipartUtil() {}

    /** Checks to see if the block at the given position currently contains a full {@link MultipartContainer}. This does
     * not take into account {@link NativeMultipart}. */
    // Left for binary compatibility
    @Nullable
    public static MultipartContainer get(class_1937 world, class_2338 pos) {
        return MultipartUtilImpl.get(world, pos);
    }

    /** Checks to see if the block at the given position currently contains a full {@link MultipartContainer}. This does
     * not take into account {@link NativeMultipart}. */
    @Nullable
    public static MultipartContainer get(class_1922 view, class_2338 pos) {
        return MultipartUtilImpl.get(view, pos);
    }

    /** Offers the given {@link AbstractPart} into the block at the given position, respecting nearby entities' bounding
     * boxes. This may return a non-null {@link PartOffer} if */
    @Nullable
    public static PartOffer offerNewPart(class_1937 world, class_2338 pos, MultipartCreator creator) {
        return offerNewPart(world, pos, creator, true);
    }

    /** Offers the given {@link AbstractPart} into the block at the given position. This may return a non-null
     * {@link PartOffer} if */
    @Nullable
    public static PartOffer offerNewPart(class_1937 world, class_2338 pos, MultipartCreator creator, boolean respectEntityBBs) {
        return MultipartUtilImpl.offerNewPart(world, pos, creator, respectEntityBBs);
    }

    /** Turns an existing {@link NativeMultipart} block into a {@link class_2586} based {@link MultipartContainer}.
     * This is only useful for updating old blocks, as adding new parts can be done via
     * {@link #offerNewPart(class_1937, class_2338, MultipartCreator)}. */
    @Nullable
    public static MultipartContainer turnIntoMultipart(class_1937 world, class_2338 pos) {
        return MultipartUtilImpl.turnIntoMultipart(world, pos);
    }
}
