/*
 * Copyright (c) 2019 AlexIIL
 *
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at https://mozilla.org/MPL/2.0/.
 */
package alexiil.mc.lib.multipart.api;

import alexiil.mc.lib.multipart.impl.LibMultiPart;
import net.minecraft.class_169;
import net.minecraft.class_176;
import net.minecraft.class_181;

/** Stores some {@link class_169}s for LMP. */
public final class PartLootParams {
    private PartLootParams() {}

    /** The primary {@link AbstractPart} that was broken */
    public static final class_169<BrokenPart> BROKEN_PART;

    /** Any additional {@link AbstractPart}'s that were broken because they
     * {@link AbstractPart#addRequiredPart(AbstractPart) required} the main broken part. */
    public static final class_169<BrokenPart[]> ADDITIONAL_PARTS;

    public static final class_176 PART_TYPE;

    static {
        BROKEN_PART = new class_169<>(LibMultiPart.id("broken_part"));
        ADDITIONAL_PARTS = new class_169<>(LibMultiPart.id("additional_parts"));
        PART_TYPE = new class_176.class_177()//
            // Block
            .method_781(class_181.field_1224)//
            .method_781(class_181.field_24424)//
            .method_781(class_181.field_1229)//
            .method_780(class_181.field_1226)//
            .method_780(class_181.field_1228)//
            .method_780(class_181.field_1225)//
            // Ours
            .method_781(BROKEN_PART)//
            .method_781(ADDITIONAL_PARTS)//
            .method_782();
    }

    /** An {@link AbstractPart} that was broken.
     * <p>
     * There are two subclasses: {@link BrokenSinglePart} (for normal parts), and {@link BrokenSubPart} (for
     * {@link SubdividedPart}s). */
    public static abstract class BrokenPart {
        private BrokenPart() {
            // Private: only BrokenSinglePart and BrokenSubPart are meant to extend this
        }

        /** @return The {@link AbstractPart} associated with this. */
        public abstract AbstractPart getPart();
    }

    public static final class BrokenSinglePart extends BrokenPart {
        public final AbstractPart part;

        public BrokenSinglePart(AbstractPart part) {
            this.part = part;
        }

        @Override
        public AbstractPart getPart() {
            return part;
        }
    }

    public static final class BrokenSubPart<Sub> extends BrokenPart {
        public final SubdividedPart<Sub> mainPart;
        public final Sub subPart;

        public BrokenSubPart(SubdividedPart<Sub> mainPart, Sub subPart) {
            if (!(mainPart instanceof AbstractPart)) {
                throw new ClassCastException(
                    mainPart.getClass() + " implements " + SubdividedPart.class + " but doesn't extend "
                        + AbstractPart.class
                );
            }
            this.mainPart = mainPart;
            this.subPart = subPart;
        }

        @Override
        public AbstractPart getPart() {
            return (AbstractPart) mainPart;
        }
    }
}
