/*
 * Copyright (c) 2019 AlexIIL
 *
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at https://mozilla.org/MPL/2.0/.
 */
package alexiil.mc.lib.multipart.impl;

import java.util.Collection;
import java.util.List;
import java.util.stream.Collectors;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;

import com.google.common.collect.ImmutableList;

import net.fabricmc.fabric.api.networking.v1.PlayerLookup;
import net.fabricmc.fabric.api.rendering.data.v1.RenderAttachmentBlockEntity;
import net.fabricmc.fabric.api.server.PlayerStream;
import net.minecraft.class_1657;
import net.minecraft.class_1937;
import net.minecraft.class_2338;
import net.minecraft.class_2415;
import net.minecraft.class_2470;
import net.minecraft.class_2487;
import net.minecraft.class_2586;
import net.minecraft.class_2680;
import net.minecraft.class_3218;
import net.minecraft.class_3222;
import alexiil.mc.lib.net.NetIdDataK;
import alexiil.mc.lib.net.NetIdDataK.IMsgDataWriterK;
import alexiil.mc.lib.net.NetIdTyped;
import alexiil.mc.lib.net.ParentNetIdSingle;
import alexiil.mc.lib.net.impl.ActiveMinecraftConnection;
import alexiil.mc.lib.net.impl.BlockEntityInitialData;
import alexiil.mc.lib.net.impl.CoreMinecraftNetUtil;
import alexiil.mc.lib.net.impl.McNetworkStack;

import alexiil.mc.lib.attributes.AttributeList;
import alexiil.mc.lib.attributes.AttributeProviderBlockEntity;

import alexiil.mc.lib.multipart.api.event.NeighbourUpdateEvent;
import alexiil.mc.lib.multipart.api.render.PartModelKey;
import alexiil.mc.lib.multipart.impl.client.PartModelData;
import alexiil.mc.lib.multipart.mixin.api.IUnloadableBlockEntity;

public class MultipartBlockEntity extends class_2586
    implements IUnloadableBlockEntity, RenderAttachmentBlockEntity, BlockEntityInitialData,
    AttributeProviderBlockEntity
{
    static final ParentNetIdSingle<MultipartBlockEntity> NET_KEY;

    static {
        NET_KEY = McNetworkStack.BLOCK_ENTITY.subType(MultipartBlockEntity.class, "libmultipart:container");
    }

    PartContainer container;

    public MultipartBlockEntity(class_2338 pos, class_2680 state) {
        super(LibMultiPart.BLOCK_ENTITY, pos, state);
        container = new PartContainer(this);
    }

    MultipartBlockEntity(PartContainer from, class_2338 pos, class_2680 state) {
        super(LibMultiPart.BLOCK_ENTITY, pos, state);
        this.container = from;
        container.blockEntity = this;
    }

    @Override
    public void method_11014(class_2487 tag) {
        super.method_11014(tag);
        if (tag.method_10545("container")) {
            container.fromTag(tag.method_10562("container"));
        }
    }

    @Override
    public class_2487 method_11007(class_2487 tag) {
        super.method_11007(tag);
        tag.method_10566("container", container.toTag());
        return tag;
    }

    @Override
    public void sendInitialData(class_3222 to) {
        PartContainer.NET_INITIAL_RENDER_DATA.send(CoreMinecraftNetUtil.getConnection(to), container);
    }

    public void applyRotation(class_2470 rotation) {
        if (rotation != class_2470.field_11467) {
            container.rotate(rotation);
        }
    }

    public void applyMirror(class_2415 mirror) {
        if (mirror != class_2415.field_11302) {
            container.mirror(mirror);
        }
    }

    @Nonnull
    class_1937 world() {
        final class_1937 w = field_11863;
        if (w != null) {
            return w;
        } else {
            throw new IllegalStateException("This doesn't have a world!");
        }
    }

    @Override
    public void method_10996() {
        super.method_10996();
        container.validate();
    }

    @Override
    public void method_11012() {
        super.method_11012();
        container.invalidate();
    }

    @Override
    public void onChunkUnload() {
        container.onChunkUnload();
    }

    public void onRemoved() {
        container.onRemoved();
    }

    public PartContainer getContainer() {
        return container;
    }

    public final boolean isServerWorld() {
        return field_11863 instanceof class_3218;
    }

    public final boolean isClientWorld() {
        return LibMultiPart.isWorldClientPredicate.test(field_11863);
    }

    public boolean canPlayerInteract(class_1657 player) {
        if (field_11863 == null || world().method_8321(field_11867) != this) {
            return false;
        }
        return player.method_5649(field_11867.method_10263() + 0.5D, field_11867.method_10264() + 0.5D, field_11867.method_10260() + 0.5D) < 64.0D;
    }

    /** @return The {@link ActiveMinecraftConnection} to use to send to the specified player. */
    public final static ActiveMinecraftConnection getPlayerConnection(class_1657 player) {
        return CoreMinecraftNetUtil.getConnection(player);
    }

    /** @return The {@link ActiveMinecraftConnection} to use to send data to the server, from the client. */
    public final ActiveMinecraftConnection getClientConnection() {
        if (!world().field_9236) {
            throw new IllegalArgumentException("We can't send data from the server to itself!");
        }
        return CoreMinecraftNetUtil.getClientConnection();
    }

    public final Collection<class_3222> getPlayersWatching() {
        return PlayerLookup.tracking(this);
    }

    /** Sends a network update update of the specified ID. */
    final <T> void sendNetworkUpdate(@Nullable class_1657 except, T obj, NetIdTyped<T> netId) {
        if (isClientWorld()) {
            netId.send(getClientConnection(), obj);
        } else if (isServerWorld()) {
            for (class_1657 player : getPlayersWatching()) {
                if (player != except) {
                    netId.send(getPlayerConnection(player), obj);
                }
            }
        }
    }

    /** Sends a network update update of the specified ID. */
    final <T> void sendNetworkUpdate(
        @Nullable class_1657 except, T obj, NetIdDataK<T> netId, IMsgDataWriterK<T> writer
    ) {

        if (isClientWorld()) {
            netId.send(getClientConnection(), obj, writer);
        } else if (isServerWorld()) {
            for (class_1657 player : getPlayersWatching()) {
                if (player != except) {
                    netId.send(getPlayerConnection(player), obj, writer);
                }
            }
        }
    }

    @Override
    public PartModelData getRenderAttachmentData() {
        ImmutableList.Builder<PartModelKey> list = ImmutableList.builder();
        for (PartHolder holder : container.parts) {
            PartModelKey key = holder.part.getModelKey();
            if (key != null) {
                list.add(key);
            }
        }
        ImmutableList<PartModelKey> built = list.build();
        // Refresh this, just to be on the safe side.
        container.partModelKeys = built;
        return new PartModelData(built);
    }

    // Events

    public void tick() {
        container.tick();
    }

    @Override
    public void addAllAttributes(AttributeList<?> list) {
        container.addAllAttributes(list);
    }

    public void onNeighbourUpdate(class_2338 otherPos) {
        container.fireEvent(new NeighbourUpdateEvent(otherPos));
    }
}
