/*
 * Copyright (c) 2019 AlexIIL
 *
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at https://mozilla.org/MPL/2.0/.
 */
package alexiil.mc.lib.multipart.mixin.api;

import javax.annotation.Nullable;

import net.fabricmc.api.EnvType;
import net.fabricmc.api.Environment;
import net.minecraft.class_1657;
import net.minecraft.class_1799;
import net.minecraft.class_1922;
import net.minecraft.class_1936;
import net.minecraft.class_1937;
import net.minecraft.class_2248;
import net.minecraft.class_2338;
import net.minecraft.class_243;
import net.minecraft.class_2586;
import net.minecraft.class_259;
import net.minecraft.class_265;
import net.minecraft.class_2680;
import net.minecraft.class_3225;

/** Interface for blocks to implement if parts of them can be targeted and broken independently from other parts. */
public interface IBlockMultipart<T> {
    Class<T> getKeyClass();

    /** Multipart version of {@link class_2248#method_9606(class_2680, class_1937, class_2338, class_1657)} */
    void onBlockBreakStart(class_2680 state, class_1937 world, class_2338 pos, class_1657 player, T subpart);

    /** Multipart version of {@link class_2248#method_9594(class_2680, class_1657, class_1922, class_2338)} */
    float calcBlockBreakingDelta(class_2680 state, class_1657 player, class_1922 view, class_2338 pos, T subpart);

    /** Multipart version of {@link class_2248#method_9576(class_1937, class_2338, class_2680, class_1657)} */
    void onBreak(class_1937 world, class_2338 pos, class_2680 state, class_1657 player, T subpart);

    /** Called instead of {@link class_1937#method_8650(class_2338, boolean)} in
     * {@link class_3225#method_14266} */
    boolean clearBlockState(class_1937 world, class_2338 pos, T subpart);

    @Environment(EnvType.CLIENT)
    default boolean playHitSound(class_1937 world, class_2338 pos, class_2680 state, class_1657 player, T subpart) {
        return false;
    }

    /** Multipart version of {@link class_2248#method_9585(class_1936, class_2338, class_2680)} */
    void onBroken(class_1936 world, class_2338 pos, class_2680 state, T subpart);

    /** Multipart version of
     * {@link class_2248#method_9556(class_1937, class_1657, class_2338, class_2680, class_2586, class_1799)}. */
    void afterBreak(
        class_1937 world, class_1657 player, class_2338 pos, class_2680 state, @Nullable class_2586 be, class_1799 stack,
        T subpart
    );

    /** @deprecated Please use {@link #getPartOutline(class_2680, class_1922, class_2338, class_243)} instead. */
    @Deprecated
    class_265 getPartOutlineShape(class_2680 state, class_1937 world, class_2338 pos, class_243 hitVec);

    default class_265 getPartOutline(class_2680 state, class_1922 view, class_2338 pos, class_243 hitVec) {
        if (view instanceof class_1937) {
            return getPartOutlineShape(state, (class_1937) view, pos, hitVec);
        } else {
            return class_259.method_1073();
        }
    }

    /** @deprecated Please use {@link #getMultipartTarget(class_2680, class_1922, class_2338, class_243)} instead. */
    @Nullable
    @Deprecated
    T getTargetedMultipart(class_2680 state, class_1937 world, class_2338 pos, class_243 hitVec);

    @Nullable
    default T getMultipartTarget(class_2680 state, class_1922 view, class_2338 pos, class_243 vec) {
        if (view instanceof class_1937) {
            return getTargetedMultipart(state, (class_1937) view, pos, vec);
        } else {
            return null;
        }
    }
}
