/*
 * Copyright (c) 2019 AlexIIL
 *
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at https://mozilla.org/MPL/2.0/.
 */
package alexiil.mc.lib.multipart.mixin.impl;

import org.spongepowered.asm.mixin.Mixin;
import org.spongepowered.asm.mixin.Shadow;
import org.spongepowered.asm.mixin.Unique;
import org.spongepowered.asm.mixin.injection.At;
import org.spongepowered.asm.mixin.injection.Inject;
import org.spongepowered.asm.mixin.injection.Redirect;
import org.spongepowered.asm.mixin.injection.callback.CallbackInfo;
import org.spongepowered.asm.mixin.injection.callback.CallbackInfoReturnable;
import alexiil.mc.lib.multipart.impl.LibMultiPart;
import alexiil.mc.lib.multipart.mixin.api.IBlockMultipart;
import net.minecraft.class_1657;
import net.minecraft.class_1799;
import net.minecraft.class_1922;
import net.minecraft.class_1937;
import net.minecraft.class_2248;
import net.minecraft.class_2338;
import net.minecraft.class_243;
import net.minecraft.class_2586;
import net.minecraft.class_265;
import net.minecraft.class_2680;
import net.minecraft.class_3218;
import net.minecraft.class_3222;
import net.minecraft.class_3225;
import net.minecraft.class_3965;

@Mixin(class_3225.class)
public class ServerPlayerInteractionManagerMixin {

    // TODO: Render the correct part outline!
    // TODO: Render the correct damage model!

    @Shadow
    class_3218 world;

    @Shadow
    class_3222 player;

    @Shadow
    boolean mining;

    @Unique
    Object multipartKey;

    /** Sent from the client by a custom break packet (as a replacement for the normal "on block start break"
     * packet). */
    @Unique
    class_243 sentHitVec;

    private void log(String text) {
        LibMultiPart.LOGGER.info("[player-interaction] '" + player.method_5820() + "' " + text);
    }

    @Inject(method = "Lnet/minecraft/server/network/ServerPlayerInteractionManager;update()V", at = @At("RETURN"))
    void update(CallbackInfo ci) {
        if (!mining && multipartKey != null) {
            if (LibMultiPart.DEBUG) {
                log("update(): Cleared multipartKey");
            }
            multipartKey = null;
        }
    }

    @Redirect(
        method = "Lnet/minecraft/server/network/ServerPlayerInteractionManager;processBlockBreakingAction("
            + "Lnet/minecraft/util/math/BlockPos;Lnet/minecraft/network/packet/c2s/play/PlayerActionC2SPacket$Action;"
            + "Lnet/minecraft/util/math/Direction;I)V",
        at = @At(value = "INVOKE",
            target = "Lnet/minecraft/block/BlockState;onBlockBreakStart(Lnet/minecraft/world/World;"
                + "Lnet/minecraft/util/math/BlockPos;Lnet/minecraft/entity/player/PlayerEntity;)V"))
    void onBlockBreakStart(class_2680 state, class_1937 w, class_2338 pos, class_1657 pl) {
        if (LibMultiPart.DEBUG) {
            log("onBlockBreakStart( " + pos + " " + state + " )");
        }
        if (state.method_26204() instanceof IBlockMultipart<?>) {
            IBlockMultipart<?> block = (IBlockMultipart<?>) state.method_26204();
            onBlockBreakStart0(state, w, pos, pl, block);
        } else {
            state.method_26179(w, pos, pl);
        }
    }

    private <T> void onBlockBreakStart0(
        class_2680 state, class_1937 w, class_2338 pos, class_1657 pl, IBlockMultipart<T> block
    ) {

        class_243 vec = sentHitVec;
        if (vec == null) {
            if (LibMultiPart.DEBUG) {
                log("onBlockBreakStart0(): vec was null!");
            }
            // Guess the hit vec from the player's look vector
            class_265 shape = state.method_26218(w, pos);
            class_3965 rayTrace = shape
                .method_1092(pl.method_5836(1), pl.method_5836(1).method_1019(pl.method_5828(1).method_1021(10)), pos);
            if (LibMultiPart.DEBUG) {
                log("onBlockBreakStart(): rayTrace = " + rayTrace);
            }
            if (rayTrace == null) {
                // This shouldn't really happen... lets just fail.
                return;
            }
            vec = rayTrace.method_17784();
        }
        if (LibMultiPart.DEBUG) {
            log("onBlockBreakStart(): vec = " + vec);
        }
        T bestKey = block.getTargetedMultipart(state, w, pos, vec);
        if (LibMultiPart.DEBUG) {
            log("onBlockBreakStart(): bestKey = " + bestKey);
        }

        if (bestKey == null) {
            // TODO!
        } else {
            this.multipartKey = bestKey;
            block.onBlockBreakStart(state, w, pos, pl, bestKey);
        }
    }

    @Redirect(at = @At(value = "INVOKE",
        target = "Lnet/minecraft/block/BlockState;calcBlockBreakingDelta(Lnet/minecraft/entity/player/PlayerEntity;"
            + "Lnet/minecraft/world/BlockView;Lnet/minecraft/util/math/BlockPos;)F"),
        method = "*")
    float calcBlockBreakingDelta(class_2680 state, class_1657 pl, class_1922 view, class_2338 pos) {
        if (LibMultiPart.DEBUG) {
            log("calcBlockBreakingDelta( " + pos + " " + state + " )");
        }
        if (state.method_26204() instanceof IBlockMultipart<?>) {
            IBlockMultipart<?> block = (IBlockMultipart<?>) state.method_26204();
            return calcBlockBreakingDelta0(block, state, pl, view, pos);
        } else {
            return state.method_26165(pl, view, pos);
        }
    }

    private <T> float calcBlockBreakingDelta0(
        IBlockMultipart<T> block, class_2680 state, class_1657 pl, class_1922 view, class_2338 pos
    ) {
        if (multipartKey == null) {
            // We haven't actually started breaking yet?
            onBlockBreakStart(state, world, pos, player);
        }

        if (multipartKey == null) {
            return state.method_26165(pl, view, pos);
        }

        if (!block.getKeyClass().isInstance(multipartKey)) {
            if (LibMultiPart.DEBUG) {
                log(
                    "calcBlockBreakingDelta0(): Wrong key " + multipartKey.getClass() + ", expected "
                        + block.getKeyClass()
                );
            }
            return state.method_26165(pl, view, pos);
        }

        T key = block.getKeyClass().cast(multipartKey);
        return block.calcBlockBreakingDelta(state, pl, view, pos, key);
    }

    @Inject(
        method = "Lnet/minecraft/server/network/ServerPlayerInteractionManager;tryBreakBlock(Lnet/minecraft/util/math/BlockPos;)Z",
        at = @At(value = "INVOKE", target = "Lnet/minecraft/block/Block;onBreak(Lnet/minecraft/world/World;"
            + "Lnet/minecraft/util/math/BlockPos;Lnet/minecraft/block/BlockState;Lnet/minecraft/entity/player/PlayerEntity;)V"),
        cancellable = true)
    void destroyBlock(class_2338 pos, CallbackInfoReturnable<Boolean> ci) {
        if (LibMultiPart.DEBUG) {
            log("destroyBlock( " + pos + " )");
        }
        class_2680 state = world.method_8320(pos);
        if (LibMultiPart.DEBUG) {
            log("destroyBlock(): multipartKey = " + multipartKey);
        }
        if (multipartKey == null) {
            // We haven't actually started breaking yet?
            onBlockBreakStart(state, world, pos, player);
        }

        class_2248 block = state.method_26204();
        if (block instanceof IBlockMultipart<?>) {
            if (LibMultiPart.DEBUG) {
                log("destroyBlock(): multipartKey = " + multipartKey);
            }
            boolean ret = destroyBlock0(pos, state, (IBlockMultipart<?>) block);
            ci.setReturnValue(ret);
        }
    }

    private <T> boolean destroyBlock0(class_2338 pos, class_2680 state, IBlockMultipart<T> block) {
        if (multipartKey == null) {
            return false;
        }
        if (!block.getKeyClass().isInstance(multipartKey)) {
            if (LibMultiPart.DEBUG) {
                log("destroyBlock0(): Wrong key " + multipartKey.getClass() + ", expected " + block.getKeyClass());
            }
            return false;
        }
        T key = block.getKeyClass().cast(multipartKey);
        class_2586 be = world.method_8321(pos);
        block.onBreak(world, pos, state, player, key);
        if (block.clearBlockState(world, pos, key)) {
            if (LibMultiPart.DEBUG) {
                log("destroyBlock0(): IBlockMultipart.clearBlockState() -> true");
            }
            block.onBroken(world, pos, state, key);
            if (!player.method_7337()) {
                class_1799 stack = player.method_6047();
                stack.method_7952(world, state, pos, player);
                block.afterBreak(world, player, pos, state, be, stack, key);
            }
            return true;
        } else {
            if (LibMultiPart.DEBUG) {
                log("destroyBlock0(): IBlockMultipart.clearBlockState() -> false");
            }
            return false;
        }
    }

    // @Redirect(
    // method = "tryBreakBlock(Lnet/minecraft/util/math/BlockPos;)Z",
    // at = @At(
    // value = "INVOKE",
    // target =
    // "Lnet/minecraft/block/Block;afterBreak(Lnet/minecraft/world/World;Lnet/minecraft/entity/player/PlayerEntity;"
    // + "Lnet/minecraft/util/math/BlockPos;Lnet/minecraft/block/BlockState;Lnet/minecraft/block/entity/BlockEntity;"
    // + "Lnet/minecraft/item/ItemStack;)V"))
    // void afterBreak(Block block, World w, PlayerEntity pl, BlockPos pos, BlockState state, BlockEntity be,
    // ItemStack stack) {
    // if (
    // !(block instanceof IBlockMultipart<?>) || !afterBreak0(
    // (IBlockMultipart<?>) block, w, pl, pos, state, be, stack
    // )
    // ) {
    // block.afterBreak(w, pl, pos, state, be, stack);
    // }
    // }

    // private <T> boolean afterBreak0(IBlockMultipart<T> block, World w, PlayerEntity pl, BlockPos pos, BlockState
    // state,
    // BlockEntity be, ItemStack stack) {
    // if (multipartKey == null) {
    // return false;
    // }
    // if (!block.getKeyClass().isInstance(multipartKey)) {
    // return false;
    // }
    // block.afterBreak(w, pl, pos, state, be, stack, block.getKeyClass().cast(multipartKey));
    // return true;
    // }
}
